#! /bin/bash

######################################################################################
#
#  JD-5602 # ahikam   # 08/02/2018 # Support command for 'set-host-config'
#                     # Adding Array with the allowed parameters
#                     # JethroAdmin set-host-config -Dparam-name=value
#                     # hdfs.kerberos.keytab, hdfs.kerberos.principal, hdfs.kerberos.reinit.interval.sec
#                     # Add Reserverd list support to create-instance (current reserved word is "host"
#
#  JD-2930 # alekseyd #  28/07/2016 # Support command `list-storage-instance'
#                     #  instance name is limited on output for 32 chars
#                     # To consider a folder inside of "$STORAGE_ROOT_PATH" an instance folder,
#                     # we check the folder has both `data' and `metaData' subfolder
#
#  JD-3097 # alekseyd #  28/07/2016 # Instance name is limited to 32 chars; old instances
#                     #  with longer name are still supported
#
#  JD-3011 # alekseyd #  28/07/2016 #  collect-logs now checks for whether instance exists,
#                     #  warns and stops executions if it doesn't
#
#  JD-3016 # alekseyd #  28/07/2016 # collect-logs now supports optional `-p <psswd>' parameter
#                     #  without password provided, 'jethro' default value will be used
#                     #  instead

#  JD-3807 # alekseyd #  15/01/2016 # collect-logs now: 1) always picks up monitor.log;
#                     #  2) stores result of SHOW CUBES;
#                     #  3) runs dmesg with timestamps where possible;
#
#  JD-3771 # alekseyd #  15/01/2016 # discovers location of service (/sbin/ vs /usr/sbin)
#                     #  and runs from there

DEBUG=FALSE
#DEBUG=TRUE
hostconf=/opt/jethro/config/host-conf.ini
MAX_NAME_LEN=32
RESERVED_WORDS=(host)
#Enable fancy colors only if JethroAdmin script onto terminal
if [[ -t 1 ]]; then
    NORMAL="\e[00m"
    GREY="\[\e[00;30m\]"
    RED="\e[00;31m"
    GREEN="\e[00;32m"
    YELLOW="\e[00;33m"
    BLUE="\e[00;34m"
    PURPLE="\[\e[00;35m\]"
    DARKGREY="\[\e[01;30m\]"
else
    unset NORMAL GREY RED GREEN YELLOW BLUE PURPLE DARKGREY
fi

#exit 9  - not all elements inserted
#exit 10 - JETHRO_HOME doesn't exists
#exit 12 - instance_name exists
#exit 13 - wrong instance_name
#exit 14 - problem with HDFS
#exit 16 - unknown storage type
#exit 18 - problem with local dir
#exit 20 - problem with cache dir
#exit 21 - need to set at least 1 parameter
#exit 50 - wrong password
#exit 99 - unexpected


function on_exit {
    echo -en ${NORMAL}
    [[ -e "$SESSION_TMP" ]] && rm -rf $SESSION_TMP
}

SESSION_TMP=`mktemp -d 2>/dev/null`
trap on_exit EXIT INT TERM

SERVICE=service
if [ -x /sbin/service ]; then
  SERVICE=/sbin/service
elif [ -x /usr/sbin/service ]; then
  SERVICES=/usr/sbin/service
fi


function display_usage {
  echo "JethroAdmin - a utility for maintaining JethroData instances "
  echo "USAGE: JethroAdmin operation instance_name [arguments...]"
  echo
  echo "   Operations"
  echo "      list-instances : list all locally attached instances"
  echo "      list-storage-instances : list all instances located under given path"
  echo "      create-instance: creates a new instance on HDFS"
  echo "      attach-instance: attaches to an existing instance on HDFS"
  echo "      detach-instance: detaches from an existing instance, keeping it on HDFS"
  echo "      delete-instance: detaches from an existing instance and deletes it from HDFS"
  echo "      change-pass:  change password"
  #echo "      check-env : post install check"
  echo -e "      collect-logs: without instance collecting only system info, with instance collect also instance logs\n"\
          "                   For instances with non-default password you will need to supply a password"
  echo "      set-host-config: configure host parameters"
  echo
  echo "   Syntax:"
  echo "      JethroAdmin list-instances"
  echo "      JethroAdmin list-storage-instances -storage-path=storage-root-path -Dstorage.type={POSIX|HDFS}"
  echo "      JethroAdmin create-instance instance-name -storage-path=storage-root-path -cache-path=cache-root-path -cache-size=cache-size<Gb> [-Dparam-name=value]"
  echo "      JethroAdmin attach-instance instance-name -storage-path=storage-root-path -cache-path=cache-root-path -cache-size=cache-size<Gb> [-Dparam-name=value]"
  echo "      JethroAdmin detach-instance instance-name"
  echo "      JethroAdmin delete-instance instance-name"
  echo "      JethroAdmin change-pass instance-name"
  #echo "      JethroAdmin check-env [st=HDFS|POSIX instance-name HDFS-user]"
  echo "      JethroAdmin collect-logs instance-name [-p <password>]"
  echo "      JethroAdmin set-host-config -Dhost-param-name=value [-Dhost-param-name=value]"
  echo "       Host-param-names:"
  echo "        -Dhdfs.kerberos.keytab={Full path to kerberos keytab}"
  echo "        -Dhdfs.kerberos.principal={username@DOMAIN}"
  echo "        -Dhdfs.kerberos.reinit.interval.sec={(optional) The renewel interval in seconds. The default value is 3600 (one hour)}"
  exit 1
}

function check_jethro_home {
  function check_that_file_exists {
    if [[ ! -r $FILE_TO_CHECK ]];then
       echo
       echo "Error - cannot locate required file $FILE_TO_CHECK, aborting operation."
       exit 10
    fi
  }

  if [[ `uname | egrep *CYG*| wc -l` -gt 0 ]]; then
    echo 'cygwin JH'
    JETHRO_HOME=`cygpath.exe -m $JETHRO_HOME`
  fi

  if [[ ! $JETHRO_HOME ]]; then
    echo
    echo "Error - \$JETHRO_HOME is not defined, aborting operation."
    exit 10
  fi

  if [[ ! -d $JETHRO_HOME ]]; then
    echo
    echo "Error - \$JETHRO_HOME does not point to an existing directory, aborting operation."
    exit 10
  fi

  if [[ $DEBUG == TRUE ]]; then echo "  DEBUG: JETHRO_HOME $JETHRO_HOME" >> /tmp/jethro_diagnostics.out 2>&1 ; fi

  if [[ ! -e $JETHRO_HOME/instances ]]; then
    echo
    echo "Error - Symbolic link \$JETHRO_HOME/instances does not exist and cannot be recreated, aborting operation."
    exit 10

  fi

  if [[ $OPERATION == create-instance ||  $OPERATION == attach-instance || $OPERATION == check-env ]]; then
   # check template files that will be copied

    FILE_TO_CHECK=$JETHRO_HOME/conf/default-jethrolog.properties
    check_that_file_exists
    FILE_TO_CHECK=$JETHRO_HOME/conf/default-local-config.ini
    check_that_file_exists
    FILE_TO_CHECK=$JETHRO_HOME/conf/default-local-cache.ini
    check_that_file_exists
    FILE_TO_CHECK=$JETHRO_HOME/conf/default-libjdmemtool-config.sh
    check_that_file_exists
  fi

}

function containsElement () {
  # returns 0 if element(first parameter) is in the list
  local e
  for e in "${@:2}"
  do
    if [[ "$e" == "$1" ]]; then
      return 0
    fi
  done
  return 1
}

# Checking if the user parameters are correct
function checkingCorrectHostConfParameters(){
   # Creating a array with all possible parameters
   hostConfArray=(hdfs.kerberos.keytab hdfs.kerberos.principal hdfs.kerberos.reinit.interval.sec)
   for parameter in $EXTRA_PARAMS
     do
       if [[ $parameter != *-D* ]]; then #-D is removed
         opt=`echo $parameter | sed -r 's/^.{1}//'`
         echo $opt
       else
         opt=`echo $parameter | sed -r 's/^.{2}//'`
       fi
       replaceToSpace=$(echo $opt | tr "=" "\n")
       temp=($replaceToSpace)
       PARAM_NAME=${temp[0]}
       PARAM_VALUE=${temp[1]}

       if  [[ $PARAM_NAME ==  "hdfs.kerberos.reinit.interval.sec" ]]; then
            # Check if the interval parameters is an Integer and it's value is bigger than 60 seconds
            if  ! [[ "$PARAM_VALUE" =~ ^[0-9]+$ ]] || [[ "$PARAM_VALUE" -lt 60 ]]  ; then
                 echo
                 echo "Parameter $PARAM_NAME should be an integer number with a minimum value of 60 Sec";
                 exit 18
            fi
       fi
       # Check if the parameter that was provided by the user is one of the possible parameters in the hostConfArray array
       inarray=$(echo ${hostConfArray[@]} | grep -o $PARAM_NAME | wc -w)
       if [ $inarray == 0 ] ; then
          echo
          echo -e "Error - Parameter $PARAM_NAME is incorrect. Please use ${hostConfArray[@]} only."
          exit 18
       fi
    done
        echo >> $hostconf
        echo "#$(date +%Y-%m-%d" "%H:%M:%S) - Modified by JethroAdmin" >> $hostconf
}


function check_extra_options {
  # loop over extra parameters
  # special handling for storage.type - it is needed during the script
  # The rest of the variables are validated and added to an array, later added to local-conf.ini
  # verifying that there are no duplicate parameters
  INI_COUNTER=0
  for i in $EXTRA_PARAMS
  do
    if [ -n "$WAITING_PASSWORD" ]; then
        unset WAITING_PASSWORD
        password=$i
        continue;
    fi

    if [[ $i != *-D* ]]; then
      opt=`echo $i | sed -r 's/^.{1}//'`
    else
      opt=`echo $i | sed -r 's/^.{2}//'`
    fi
    arr=$(echo $opt | tr "=" "\n")
    temp=($arr)
    PARAM_NAME=${temp[0]}
    PARAM_VALUE=${temp[1]}
    case $PARAM_NAME in
      storage.type)
        STORAGE_TYPE=`echo ${PARAM_VALUE} | tr [:lower:] [:upper:]`
        if [[ $STORAGE_TYPE != "HDFS" && $STORAGE_TYPE != "POSIX" ]]; then
          echo
          echo -e "Error - unknown value ${RED}<$STORAGE_TYPE>${NORMAL} for parameter storage.type, aborting operation."
          exit 18
        fi
        ;;

      # Insert parameters into host-conf.ini
      hdfs.kerberos.keytab)
      echo $PARAM_NAME=$PARAM_VALUE >> $hostconf
      ;;
      hdfs.kerberos.principal)
         echo $PARAM_NAME=$PARAM_VALUE >> $hostconf
      ;;
      hdfs.kerberos.reinit.interval.sec)
         echo $PARAM_NAME=$PARAM_VALUE >> $hostconf
      ;;


      # All other parameters will be inserted into a vector and added to the local-config.ini

      storage-path)

        STORAGE_ROOT_PATH="$PARAM_VALUE"

        ;;

      cache-path)
        if [[ ! -d $PARAM_VALUE ]]; then
          echo -e "Error - cache-path directory ${RED}<$PARAM_VALUE>${NORMAL} does not exist, aborting operation."
          exit 18
        fi
        if [[ ! -w $PARAM_VALUE ]]; then
          echo -e "Error -  cache-path directory ${RED}<$PARAM_VALUE>${NORMAL} is not writable, aborting operation."
          exit 18
        fi

        CACHE_ROOT_PATH=$PARAM_VALUE
        CACHE_PATH=$CACHE_ROOT_PATH/$INSTANCE_NAME
        #mkdir $CACHE_PATH/adaptive

        TOTAL_SIZE=`df -Pk $PARAM_VALUE | tail -1 |  awk '{print $4}'`
        # translate user input to absolute path
        PARAM_VALUE=$(readlink -f "$PARAM_VALUE")

        containsElement $PARAM_NAME "${INI_PARAM_NAMES[@]}"
        if [[ $? -eq 0 ]]; then
          echo -e "Error - parameter ${RED}<$PARAM_NAME>${NORMAL} appears twice, aborting operation."
          exit 18
        else
          INI_PARAM_NAMES[$INI_COUNTER]="local.cache.path"
          INI_PARAMS[$INI_COUNTER]="local.cache.path"\=$CACHE_ROOT_PATH
          INI_COUNTER=$((INI_COUNTER + 1))
          #INI_PARAM_NAMES[$INI_COUNTER]="adaptive.cache.location"
          #INI_PARAMS[$INI_COUNTER]="adaptive.cache.location"\=$CACHE_PATH/adaptive
          #INI_COUNTER=$((INI_COUNTER + 1))
        fi
        ;;

      cache-size)
        if [[ ! $PARAM_VALUE =~ ^[0-9]+(g|G|Gb|gb)?$ ]]; then
          echo -e "Error - parameter size ${RED}<$PARAM_VALUE>${NORMAL} is not a valid size, aborting operation."
          exit 18
        fi

        CACHE_SIZE=${PARAM_VALUE//[!0-9]/}
        let cache_size_bytes=$CACHE_SIZE*1000000
        if [[ $cache_size_bytes -gt $TOTAL_SIZE ]]; then
          echo -e "ERROR - parameter cache-size ${RED}<$PARAM_VALUE>${NORMAL} is greater than available free space <$TOTAL_SIZE> at mount point <$CACHE_PATH>, aborting operation."
          exit 18
        fi

        #set local cache size
        if [[ `uname | egrep *CYG*| wc -l` -gt 0 ]]; then
          CACHE_SIZE_LOCAL=0
        else
          CACHE_SIZE_LOCAL=$CACHE_SIZE
        fi

        containsElement $PARAM_NAME "${INI_PARAM_NAMES[@]}"
        if [[ $? -eq 0 ]]; then
          echo -e "ERROR - parameter ${RED}<$PARAM_NAME>${NORMAL} appears twice, aborting operation."
          exit 18
        else
          INI_PARAM_NAMES[$INI_COUNTER]="local.cache.max.size"
          INI_PARAMS[$INI_COUNTER]="local.cache.max.size"\="$CACHE_SIZE_LOCAL"Gb
          INI_COUNTER=$((INI_COUNTER + 1))
        fi
        ;;
      p)
        if [ -z "$PARAM_VALUE" ]; then
          WAITING_PASSWORD=1
        else
          password=$PARAM_VALUE
        fi
        ;;
      *)
        echo "ERROR - unknown command line parameter $i. Please use -D[paramter] to insert parameter. aborting operation."
        exit 1
        ;;
    esac
    #fi
  done

}

function check_local_dir {
  function check_free_space {
      FSIZE=`df -Pk $FS | tail -1 |  awk '{print $4}'`
      if [[ $FSIZE -lt 1048576 ]]; then
        echo "Error - less than 1GB free in $FS, aborting operation."
        exit 1
      fi
      if [[ $FSIZE -lt 5242880 ]]; then
        echo "Warning - less than 5GB available in $FS. It is recommended to free up some space post-install"
      fi
  }

  if [[ ! -d $JETHRO_HOME/instances ]]; then
    echo
    echo "Error - local instances directory $JETHRO_HOME/instances does not exist, aborting operation."
    exit 18
  fi

  if [[ ! -w $JETHRO_HOME/instances ]]; then
    echo
    echo "Error - local instances directory $JETHRO_HOME/instances is not writable, aborting operation."
    exit 18
  fi
  
  if [[ ! -w /var/log/jethro ]]; then
    echo
    echo "Error - /var/log/jetho is not writable, aborting operation."
    exit 18
  fi

  if [[ ! -w /var/run/jethro ]]; then
    echo
    echo "Error - /var/run/jethro is not writable, aborting operation."
    exit 18
  fi


  if [[ -d $JETHRO_HOME/instances/$INSTANCE_NAME ]]; then
    LOCAL_DIR_EXISTS="TRUE"
  else
    LOCAL_DIR_EXISTS="FALSE"
  fi
  if [[ $DEBUG == TRUE ]]; then echo "  DEBUG: LOCAL_DIR_EXISTS $LOCAL_DIR_EXISTS" >> /tmp/jethro_diagnostics.out 2>&1 ; fi

  # check free space if operation will create directory on linux
  if [[ $OPERATION == create-instance ||  $OPERATION == attach-instance || $OPERATION == check-env ]]; then
    if [[ `uname | egrep *CYG*| wc -l` -eq 0 ]]; then
      FS=/var/log
      check_free_space
      FS=/opt
      check_free_space
    fi
  fi
}

function parse_local_config {
  LOCAL_CONFIG=$JETHRO_HOME/instances/$INSTANCE_NAME/local-conf.ini
  if [[ !  -r "$LOCAL_CONFIG" ]]; then
    echo
    echo "Error - local configuration file $LOCAL_CONFIG does not exist, aborting operation."
    exit 18
  fi
   ################################################
   # extracting STORAGE_TYPE - either HDFS or POSIX
   # values are converted to upper case
   ################################################
   STORAGE_TYPE_LINE=`grep -i storage.type $LOCAL_CONFIG | grep -v '^#'`
   if [ -z "$STORAGE_TYPE_LINE" ]; then
     echo
     echo "Error - local configuration file $LOCAL_CONFIG does not include mandatory parameter storage.type, aborting operation."
     exit 18
   fi

   STORAGE_TYPE_CLEANED=$(echo $STORAGE_TYPE_LINE | tr "=" "\n")
   STORAGE_TYPE_ARR=($STORAGE_TYPE_CLEANED)
   STORAGE_TYPE=`echo ${STORAGE_TYPE_ARR[1]} | tr [:lower:] [:upper:]`

   if [[ $STORAGE_TYPE != "HDFS" && $STORAGE_TYPE != "POSIX" ]]; then
     echo
     echo -e "Error - local configuration file $LOCAL_CONFIG has an unknown value ${RED}<$STORAGE_TYPE>${NORMAL} for parameter storage.type, aborting operation."
     exit 18
   fi
   if [[ $DEBUG == TRUE ]]; then echo "  DEBUG: STORAGE_TYPE $STORAGE_TYPE" >> /tmp/jethro_diagnostics.out 2>&1 ; fi

   #########################
   # extracting STORAGE_PATH
   #########################
   STORAGE_PATH_LINE=`grep -i 'storage.root.path' $LOCAL_CONFIG | grep -v '^#'`
   if [ -z "$STORAGE_PATH_LINE" ]; then
     echo
     echo "Error - local configuration file $LOCAL_CONFIG does not include mandatory parameter storage.root.path, aborting operation."
     exit 18
   fi

   STORAGE_PATH_CLEANED=$(echo $STORAGE_PATH_LINE | tr "=" "\n")
   STORAGE_PATH_ARR=($STORAGE_PATH_CLEANED)
   INSTANCE_STORAGE_PATH=${STORAGE_PATH_ARR[1]}
   STORAGE_ROOT_PATH=`dirname $INSTANCE_STORAGE_PATH`
   if [[ $DEBUG == TRUE ]]; then echo "  DEBUG: INSTANCE_STORAGE_PATH $INSTANCE_STORAGE_PATH" >> /tmp/jethro_diagnostics.out 2>&1 ; fi

   #############################
   # extracting LOCAL CACHE PATH
   #############################
   LOCAL_CACHE_LINE=`grep -i  'local.cache.path' $LOCAL_CONFIG | grep -v '^#'`
   LOCAL_CACHE_CLEANED=$(echo $LOCAL_CACHE_LINE | tr "=" "\n")
   LOCAL_CACHE_ARR=($LOCAL_CACHE_CLEANED)
   CACHE_ROOT_PATH=${LOCAL_CACHE_ARR[1]}
   CACHE_PATH=$CACHE_ROOT_PATH/$INSTANCE_NAME

   # stripping CACHE_PATH from trailing whitspace like ^M,
   #   or else it breaks string concatenation when building dir/instance path
   #CACHE_PATH="${CACHE_PATH%"${CACHE_PATH##*[![:space:]]}"}"

   if [[ $DEBUG == TRUE ]]; then echo "  DEBUG: CACHE_PATH $CACHE_PATH" >> /tmp/jethro_diagnostics.out 2>&1 ; fi

   #################################
   # extracting LOCAL CACHE MAX_SIZE
   #################################
   LOCAL_CACHE_LINE=`grep -i  'local.cache.max.size' $LOCAL_CONFIG | grep -v '^#'`
   LOCAL_CACHE_CLEANED=$(echo $LOCAL_CACHE_LINE | tr "=" "\n")
   LOCAL_CACHE_ARR=($LOCAL_CACHE_CLEANED)
   LOCAL_CACHE_MAX_SIZE=${LOCAL_CACHE_ARR[1]}
   if [[ $DEBUG == TRUE ]]; then echo "  DEBUG: LOCAL_CACHE_MAX_SIZE $LOCAL_CACHE_MAX_SIZE" >> /tmp/jethro_diagnostics.out 2>&1 ; fi
}

function check_hdfs_root {

  if [ ! -x "$(type -p hadoop)" ]; then
    echo
    echo 'ERROR - could not find "hadoop" executable in the path, EXITING'
    exit 14
  fi

  # first, check basic HDFS connectivity
  echo -n "Checking HDFS path"
  hadoop fs -ls / >/dev/null 2>&1
  RC=$?
  if [[ RC -ne 0 ]]; then
    echo
    echo "ERROR - hadoop connectivity failed, could not list files under /, aborting operation."
    exit 14
  fi

  # check if HDFS directory exists
  echo -n "."
  hadoop fs -test -d $STORAGE_ROOT_PATH  >/dev/null 2>&1
  RC=$?
  if [[ RC -ne 0 ]]; then
    echo
    echo -e "ERROR - HDFS directory ${RED}$STORAGE_ROOT_PATH${NORMAL} does not exist, aborting operation."
    exit 14
  fi
}

function check_hdfs_dir {
  check_hdfs_root
  # check if HDFS directory is writable
  echo -n "."
  TEMP_FILE=$STORAGE_ROOT_PATH/temp.test_file.jethro
  hadoop fs -test -e $TEMP_FILE  >/dev/null 2>&1
  TEMP_RC=$?
  echo -n "."
  if [[ $TEMP_RC -eq 1 ]]; then
    # file does not exist - test write access using touchz, then delete the file
    hadoop fs -touchz $TEMP_FILE >/dev/null 2>&1
    RC=$?
    echo -n "."
    hadoop fs -rm -skipTrash $TEMP_FILE >/dev/null 2>&1
  else
    # file exists, test write access using rm
    hadoop fs -rm -skipTrash $TEMP_FILE >/dev/null 2>&1
    RC=$?
  fi
  if [[ $RC -ne 0 ]]; then
    echo
    echo "ERROR - no write permission in HDFS directory $STORAGE_ROOT_PATH, aborting operation."
    exit 14
  fi

  # check if instance dir exist, return as flag
  echo -n "."

  hadoop fs -ls $INSTANCE_STORAGE_PATH > /dev/null 2>&1
  RC=$?
  if [[ $RC -ne 0 ]]; then
    HDFS_DIR_EXISTS="FALSE"
  else
    HDFS_DIR_EXISTS="TRUE"
  fi
  echo " Done"

  if [[ $DEBUG == TRUE ]]; then echo "  DEBUG: HDFS_DIR_EXISTS $HDFS_DIR_EXISTS" >> /tmp/jethro_diagnostics.out 2>&1 ; fi
}

function check_posix_root {
  # check posix root dir exists
  if [[ ! -d "$STORAGE_ROOT_PATH" ]]; then
    echo
    echo -e "ERROR - POSIX directory ${RED}$STORAGE_ROOT_PATH${NORMAL} does not exist, aborting operation."
    exit 14
  fi

}

function check_posix_dir {
  # check posix dir exists
  if [[ -d "$INSTANCE_STORAGE_PATH" ]]; then
    POSIX_DIR_EXISTS="TRUE"
  else
    POSIX_DIR_EXISTS="FALSE"
  fi

  if [[ $DEBUG == TRUE ]]; then echo "  DEBUG: POSIX_DIR_EXISTS $POSIX_DIR_EXISTS" >> /tmp/jethro_diagnostics.out 2>&1 ; fi
}


function delete_hdfs_instance {
  echo "Deleting instance from HDFS"
  hadoop fs -rm -r -skipTrash $INSTANCE_STORAGE_PATH >/dev/null 2>&1
  # RC=$?
  # if [[ $RUN -eq 0 ]]; then

}

function delete_posix_instance {
  echo "Deleting instance from local storage"
  rm -rf $INSTANCE_STORAGE_PATH >/dev/null 2>&1
  # RC=$?
  # if [[ $RUN -eq 0 ]]; then

}

function reset_hdfs_instance {
  echo -n "Resetting instance directories on HDFS"
  hadoop fs -rm -r -skipTrash $INSTANCE_STORAGE_PATH/data/* >/dev/null 2>&1
  echo -n "."
  hadoop fs -rm -r -skipTrash $INSTANCE_STORAGE_PATH/metaData/* >/dev/null 2>&1
  echo -n "."
  hadoop fs -rm -r -skipTrash $INSTANCE_STORAGE_PATH/tmp/* >/dev/null 2>&1
  echo -n "."
  hadoop fs -rm -r -skipTrash $INSTANCE_STORAGE_PATH/sys/* >/dev/null 2>&1
  echo -n "."
  hadoop fs -rm -r -skipTrash $INSTANCE_STORAGE_PATH/adaptive/* >/dev/null 2>&1
  echo -n "."
  hadoop fs -rm -r -skipTrash $INSTANCE_STORAGE_PATH/trash/* >/dev/null 2>&1
  echo " Done"
}

function reset_posix_instance {
  echo "Resetting instance directories on local storage"
  if [ "$(ls $INSTANCE_STORAGE_PATH/data/)"     ]; then rm -rf $INSTANCE_STORAGE_PATH/data/*     >/dev/null 2>&1; fi
  if [ "$(ls $INSTANCE_STORAGE_PATH/sys/)"      ]; then rm -rf $INSTANCE_STORAGE_PATH/sys/*      >/dev/null 2>&1; fi
  if [ "$(ls $INSTANCE_STORAGE_PATH/tmp/)"      ]; then rm -rf $INSTANCE_STORAGE_PATH/tmp/*      >/dev/null 2>&1; fi
  if [ "$(ls $INSTANCE_STORAGE_PATH/metaData/)" ]; then rm -rf $INSTANCE_STORAGE_PATH/metaData/* >/dev/null 2>&1; fi
  if [ "$(ls $INSTANCE_STORAGE_PATH/trash/)"    ]; then rm -rf $INSTANCE_STORAGE_PATH/trash/*    >/dev/null 2>&1; fi
  if [ "$(ls $INSTANCE_STORAGE_PATH/adaptive/)" ]; then rm -rf $INSTANCE_STORAGE_PATH/adaptive/* >/dev/null 2>&1; fi

}

function delete_local_instance {
  echo 'Stopping instance services locally, if running'
  # Stopping all services of the instance $INSTANCE_NAME
  $JETHRO_HOME/bin/jethro.initd stop $INSTANCE_NAME all
  delete_cache

  echo "Deleting local instance directory"
  rm -rf $JETHRO_HOME/instances/$INSTANCE_NAME >/dev/null
  RC=$?
  if [[ $RC -ne 0 ]]; then
    echo "  Got unexpected return code $RC from previous command 1, continuing"
  fi

  echo "Deleting local instance log directory"
  rm -rf /var/log/jethro/$INSTANCE_NAME  >/dev/null
  RC=$?
  if [[ $RC -ne 0 ]]; then
    echo "  Got unexpected return code $RC from previous command 2, continuing"
  fi

  echo "Removing instance from \$JETHRO_HOME/instances/services.ini"
  sed -i /^${INSTANCE_NAME}:/d $JETHRO_HOME/instances/services.ini >/dev/null

  RC=$?
  if [[ $RC -ne 0 ]]; then
    echo "  Got unexpected return code $RC from previous command 3, continuing"
  fi
}

function delete_cache {
  if  [ -z "$CACHE_PATH" ]; then
    echo "Skipping - cache is not defined"
  else
    echo "Deleting cache directory"
    rm -rf $CACHE_PATH/  > /dev/null 2>&1
  fi
}

function create_local_instance {
  echo "Creating local instance files"

 # check_local_dir verified that $JETHRO_HOME/instances exists, writable, no $INSTANCE_NAME sub-dir
  mkdir $JETHRO_HOME/instances/$INSTANCE_NAME
  mkdir $JETHRO_HOME/instances/$INSTANCE_NAME/tmp

  # check_jethro_home verified that the four source files exist
  cp $JETHRO_HOME/conf/default-jethrolog.properties     $JETHRO_HOME/instances/$INSTANCE_NAME/jethrolog.properties
  cp $JETHRO_HOME/conf/default-log4j.properties         $JETHRO_HOME/instances/$INSTANCE_NAME/log4j.properties
  cp $JETHRO_HOME/conf/default-local-config.ini         $JETHRO_HOME/instances/$INSTANCE_NAME/local-conf.ini
  cp $JETHRO_HOME/conf/default-local-cache.ini          $JETHRO_HOME/instances/$INSTANCE_NAME/local-cache.ini
  cp $JETHRO_HOME/conf/default-libjdmemtool-config.sh   $JETHRO_HOME/instances/$INSTANCE_NAME/libjdmemtool-config.sh

  if [[ -d /var/log/jethro ]]; then
    mkdir /var/log/jethro/$INSTANCE_NAME
    ln -sf /var/log/jethro/$INSTANCE_NAME $JETHRO_HOME/instances/$INSTANCE_NAME/log
  else # used in sanity on dev server
    echo "Warning - /var/log/jethro does not exist, creating the log dir without a symbolic link"
    mkdir $JETHRO_HOME/instances/$INSTANCE_NAME/log
  fi

  LOCAL_CONFIG=$JETHRO_HOME/instances/$INSTANCE_NAME/local-conf.ini
  echo '' >> $LOCAL_CONFIG
  echo "# ADDED BY JethroAdmin" >> $LOCAL_CONFIG
  echo "storage.type=$STORAGE_TYPE" >> $LOCAL_CONFIG
  if [[ `uname | egrep *CYG*| wc -l` -gt 0 ]]; then
    # experimental support for Windows
    echo "storage.root.path=`cygpath.exe -m $INSTANCE_STORAGE_PATH`" >> $LOCAL_CONFIG
  else
    echo "storage.root.path=$INSTANCE_STORAGE_PATH" >> $LOCAL_CONFIG

    # services.ini handling
    SERVICES_INI=$JETHRO_HOME/instances/services.ini
    # find all ports mentioned in services.ini - then find the first one from 9111 not mentioned
    PORT_LIST_CLEANED=`cut -d":" -f2 $SERVICES_INI 2> /dev/null | grep -e '^[0-9]\+$' | sort`
    PORT_LIST_ARR=($PORT_LIST_CLEANED)
    for CANDIDATE_PORT in {9111..9999}
    do
      containsElement $CANDIDATE_PORT "${PORT_LIST_ARR[@]}"
      if [[ $? -eq 1 ]]; then
        ACTUAL_PORT=$CANDIDATE_PORT
        break
      fi
    done
    echo "Adding instance to \$JETHRO_HOME/instances/services.ini, will use port $ACTUAL_PORT"
    # AUTOSTART_MAINT_SCHEDULER logic - if it was create-instance then "yes", if attach-instance then "no" to avoid multiple maint and scheduler proceses
    echo "$INSTANCE_NAME:$ACTUAL_PORT:yes:$AUTOSTART_MAINT_SCHEDULER:$AUTOSTART_MAINT_SCHEDULER" >> $SERVICES_INI
  fi

  # add additional parameters from command line
  mkdir $CACHE_PATH
  if [[ $? -ne 0 ]]; then
    echo "ERROR, cannot create cache directory $CACHE_PATH, aborting operation."
    delete_local_instance
    exit 20
  fi

  TMP=${#INI_PARAMS[@]}
  for i in `seq -s" " 0 $((TMP - 1))`
    do
      echo ${INI_PARAMS[$i]}  >> $LOCAL_CONFIG
  done
}

function print_autostart_message {
  if [[ $SERVICE_AUTOSTART_DISABLED -eq 1 ]]; then
      echo
      echo "AUTO-START DISABLED FOR THIS INSTANCE"
      echo "Auto-start of services for this instance is disabled, since there is another instance with auto-start services."
      echo "Edit \$JETHRO_HOME/instances/services.ini to modify auto-start behaviour"
      echo
  fi
}

function create_hdfs_instance {
  echo -n "Creating directories on HDFS"
  hadoop fs -mkdir "$INSTANCE_STORAGE_PATH"  2>/dev/null 1>&2
  RC=$?
  if [[ $RC -ne 0 ]]; then
    echo "  Got unexpected return code $RC while creating a directory on HDFS, cleaning up and aborting operation."
    delete_local_instance
    exit 99
  fi
   echo -n "."
   hadoop fs -mkdir $INSTANCE_STORAGE_PATH/data  2>/dev/null 1>&2
   echo -n "."
   hadoop fs -mkdir $INSTANCE_STORAGE_PATH/metaData  2>/dev/null 1>&2
   echo -n "."
   hadoop fs -mkdir $INSTANCE_STORAGE_PATH/tmp  2>/dev/null 1>&2
   echo -n "."
   hadoop fs -mkdir $INSTANCE_STORAGE_PATH/sys  2>/dev/null 1>&2
   echo -n "."
   hadoop fs -mkdir $INSTANCE_STORAGE_PATH/adaptive  2>/dev/null 1>&2
   echo -n "."
   hadoop fs -mkdir $INSTANCE_STORAGE_PATH/trash  2>/dev/null 1>&2
  echo "  Done"
}

function create_system_tables {
  echo  "Creating system tables..."
  JethroClient $INSTANCE_NAME localhost:$ACTUAL_PORT -p jethro -u jethro -i $JETHRO_HOME/bin/system_tables.ddl >/dev/null 2>&1
  #Jethro Q $INSTANCE_NAME $JETHRO_HOME/bin/system_tables.ddl >/dev/null 2>&1 #JD-5044: changed to cli call
  #TBD - verify that queries finished successfully
  echo "  Done"
  }

 function create_schemas { #JD-4946: create sys built-in schema
  echo  "Creating basic schemas..."
   JethroClient $INSTANCE_NAME localhost:$ACTUAL_PORT -p jethro -u jethro -i $JETHRO_HOME/bin/basic_schema.ddl >/dev/null 2>&1
  #TBD - verify that queries finished successfully
  echo "  Done"
  }
function create_posix_instance {
  echo  "Creating storage directories locally..."

  #mkdir $JETHRO_HOME/instances/$INSTANCE_NAME 2>/dev/null

  mkdir $INSTANCE_STORAGE_PATH 2>/dev/null
  RC=$?
  if [[ $RC -ne 0 ]]; then
    echo " Got unexpected return code $RC while creating a directory on POSIX, cleaning up and aborting operation."
    delete_local_instance
    exit 99
  fi

  mkdir $INSTANCE_STORAGE_PATH/data
  mkdir $INSTANCE_STORAGE_PATH/metaData
  mkdir $INSTANCE_STORAGE_PATH/sys
  mkdir $INSTANCE_STORAGE_PATH/tmp
  mkdir $INSTANCE_STORAGE_PATH/adaptive
  mkdir $INSTANCE_STORAGE_PATH/trash
}

function generate_attach_statement {
  if [[ $OPERATION == "create-instance" ]]; then
    # if we came from create-instance. we just paste all command line parameters from the second one
    COMMAND="JethroAdmin attach-instance ${ARGS[@]:1}"
  else
    # Else, need to get storage parameters and cache paramters from config file
    COMMAND="JethroAdmin attach-instance $INSTANCE_NAME -storage-path=$STORAGE_ROOT_PATH -cache-path=$CACHE_ROOT_PATH -cache-size=$LOCAL_CACHE_MAX_SIZE"
    if [[ $STORAGE_TYPE == "POSIX" ]]; then
      COMMAND="$COMMAND -Dstorage.type=POSIX"
    fi
    # if [[ ! -z $CACHE_PATH && ! -z $LOCAL_CACHE_MAX_SIZE ]]; then
    #   COMMAND="$COMMAND -cache=$CACHE_PATH -size=$LOCAL_CACHE_MAX_SIZE"
    # fi
  fi
  echo $COMMAND
}


#declaring ${LOCAL_LIST} as associate array
#It's declared outside the function since older versions of bash
#does not have the "global" notion
#We will use get_local_instances as both iteratable container and set
declare -A LOCAL_LIST
function get_local_instances {
    for name in $JETHRO_HOME/instances/*/
    do
        if [ -f "$name/local-conf.ini" ]; then
            local inst=`basename ${name}`   #makes ${inst} visible inside the func only
            LOCAL_LIST[$inst]=$inst
        fi
    done
}

function get_hdfs_instances {
    HDFS_INSTANCES_LIST=()  #defined as regular array:
                            #allows to iterate with 'for' and obtain num of elements

    declare -Ai ddd    #assoc array of integers, to count encounters of potential names of instances
    for name in $(hadoop fs -ls -d "$STORAGE_ROOT_PATH/*/metaData" "$STORAGE_ROOT_PATH/*/data" 2>/dev/null | awk -v FS=/  '/^d.*\/.*\/data|metaData/{print $(NF-1)}')
                 #`hadoop' command is very slow, so we tried to minimize amount of times we call it as much as possible;
                 # indeed we call it only ONCE!
                 # To consider a folder inside of "$STORAGE_ROOT_PATH" an instance folder,
                 # this command should list its name twice (once from `data' subfolder,
                 # and second time from `metaData' subfolder
    do
        ddd[$name]+=1   #count every occurance of folder 'name'
    done

    for name in "${!ddd[@]}"; do    #run on keys of array ddd
    # for instance folder counter should be exactly 2 (has both 'data' and 'metaData' subfolders)
        test ${ddd[$name]} -eq 2 && HDFS_INSTANCES_LIST+=(${name})
    done
}

function get_posix_instances {
    POSIX_INSTANCES_LIST=() #defined as regular array:
                            #allows to iterate with 'for' and obtain num of elements
    for name in $STORAGE_ROOT_PATH/*/
    do
        if [ -d "$name/data" -a -d "$name/metaData" ]; then
            POSIX_INSTANCES_LIST+=(`basename ${name}`)
        fi
    done
}

function print_unattached_hdfs_instances {
    get_local_instances
    get_hdfs_instances

    unset CANDIDATE_INSTANCES_LIST
    for x in ${HDFS_INSTANCES_LIST[@]}; do
        containsElement $x "${LOCAL_LIST[@]}"
        if [ $? -ne 0 ]; then
            CANDIDATE_INSTANCES_LIST+=(${x})
        fi
    done
    if [ -n "$CANDIDATE_INSTANCES_LIST" ]; then
        echo -e ${YELLOW}"There are some other unattached instances under $STORAGE_ROOT_PATH:"${GREEN}
        for x in ${CANDIDATE_INSTANCES_LIST[@]}; do
            echo -e "  " $x
        done
    fi
    echo -en $NORMAL
}

function print_unattached_posix_instances {
    get_local_instances
    get_posix_instances
    unset CANDIDATE_INSTANCES_LIST
    for x in ${POSIX_INSTANCES_LIST[@]}; do
       containsElement $x "${LOCAL_LIST[@]}"
       if [ $? -ne 0 ]; then
          CANDIDATE_INSTANCES_LIST+=(${x})
       fi
    done
    if [ -n "$CANDIDATE_INSTANCES_LIST" ]; then
      echo -e ${YELLOW}"There are some other unattached instances on local storage $STORAGE_ROOT_PATH:"${GREEN}
      for x in ${CANDIDATE_INSTANCES_LIST[@]}; do
          echo -e "  " $x
      done
    fi
    echo -en $NORMAL
}

###############################
# MAIN SCRIPT STARTS HERE
###############################
ARGS=( "$@" )
# General checks
if [[ $EUID -eq 0 ]]; then
  # Running as root messes local file permissions
  echo
  echo 'Error - this script must run as a non-root user, exiting'
  exit 1
fi

if [ $# -gt 0 ]; then
  OPERATION=`echo $1 | tr [:upper:] [:lower:]`
  if [[ $OPERATION != list-instances  &&
        $OPERATION != list-storage-instances &&
        $OPERATION != create-instance &&
        $OPERATION != attach-instance &&
        $OPERATION != detach-instance &&
        $OPERATION != delete-instance &&
        $OPERATION != reset-instance &&
        $OPERATION != set-host-config &&
        $OPERATION != check-env &&
        $OPERATION != collect-logs &&
        $OPERATION != change-pass
      ]]; then
    echo -e "Error - unknown operation ${RED}<$OPERATION>${NORMAL}, aborting operation."
    echo "Valid operations: list-instances, list-storage-instances, create-instance, attach-instance, detach-instance, delete-instance, change-pass, collect-logs, set-host-config"
    exit 1
  fi
else
  display_usage
fi

if  [[ $# -lt 2 ]]; then
  case $OPERATION in
   list-instances) ;;
   collect-logs)
    echo -e "Please provide ${YELLOW}instance-name${NORMAL} to collect logs"
    exit 1 ;;

   check-env) ;;
   *)
    display_usage;;
  esac
fi

if [[ "$2" != "-storage-path*" ]]; then
  INSTANCE_NAME=`echo $2 | tr [:upper:] [:lower:]`
else
  echo -e "Error, invalid instance name ${RED}<$2>${NORMAL}, aborting operation."
  exit 9
fi

if [[ $OPERATION == "collect-logs" ]]; then
  password=jethro
  EXTRA_PARAMS=${@:3}
  check_extra_options
  #################
  # Collect logs
  #################
  function noSqlStats_YesNo {
     while read -p "Would you like to collect logs WITHOUT SQL statistics [yes/no]: " ANSWER; do
       ANSWER=`echo $ANSWER | tr [:upper:] [:lower:] `
       case $ANSWER in
           yes)
              rm -f  $SESSION_TMP/sqlstatistics.out >/dev/null 2>&1
              break
              ;;
           no)
              echo "Exiting..."
              exit 50
             ;;
           *)
           continue
           ;;
       esac
     done
  }

  DEBUG=TRUE
# date format:
# %F     full date; same as %Y-%m-%d
# %R     24-hour hour and minute; same as %H:%M

  NOW=`date +"%FT%0H-%0M" 2>/dev/null`
  JETHRO_TAR=jethro_diagnostics.$INSTANCE_NAME.$NOW.tar
  umask 066    # all created files are accessible only by us
  jethro_diagnostics >> /tmp/jethro_diagnostics.out 2>&1

  dmesg -T > $SESSION_TMP/dmesg.out 2>&1
  if [ $? -ne 0 ]; then
    dmesg > $SESSION_TMP/dmesg.out 2>&1
  fi

  crontab -l >$SESSION_TMP/crontab.out 2>&1
  
  cat $JETHRO_HOME/instances/services.ini >$SESSION_TMP/services.ini 2>&1
  
  cd $SESSION_TMP
  if [[ $INSTANCE_NAME == '_all' ]]; then
    LOGS=/var/log/jethro/
    tar -cvhf $JETHRO_TAR -C /tmp jethro_diagnostics.out  -C $SESSION_TMP dmesg.out crontab.out services.ini -C $LOGS . > /dev/null 2>&1
  else
    check_local_dir
    if [[ $LOCAL_DIR_EXISTS != "TRUE" ]]; then
      echo
      echo -e "Error - instance ${RED}<$INSTANCE_NAME>${NORMAL} does not exist, aborting operation."
      echo "Reason - could not find local directory $JETHRO_HOME/instances/$INSTANCE_NAME"
      exit 12
    fi

    LOGS=/var/log/jethro/
    CONF=$JETHRO_HOME/instances/
    port="$(cat $JETHRO_HOME/instances/services.ini | grep "^$INSTANCE_NAME:" | awk -F ":" '{print $2}')"
    port=${port:-XXXX}

    if [[ "$port" =~ ^[0-9]+$ && $port -gt 1023 && $port -le 65535 ]]; then
        # $port is a number and is in range 1023 .. 65535
        JethroClient $INSTANCE_NAME localhost:"$port" -p $password  -i $JETHRO_HOME/bin/sqlstatistics.sql > $SESSION_TMP/sqlstatistics.out 2>&1
        if grep --quiet -i "invalid connect string" $SESSION_TMP/sqlstatistics.out;  then
            echo -e "Service for instance ${RED}<$INSTANCE_NAME>${NORMAL} has not been properly registered."
            echo 'Could not get SQL statistics.'
            noSqlStats_YesNo
        elif grep --quiet "Couldn't connect to any host" $SESSION_TMP/sqlstatistics.out; then
            echo -e "Server is down. Could not get SQL statistics for instance ${RED}<$INSTANCE_NAME>${NORMAL}."
            noSqlStats_YesNo
        elif grep --quiet "Authentication failed, check username and password" $SESSION_TMP/sqlstatistics.out; then
            echo -e "Connection to server has been rejected - username or password is not valid for instance ${RED}<$INSTANCE_NAME>${NORMAL}!"
            echo 'Could not get SQL statistics...'
            echo -e "${YELLOW}Please use -p <password> option of 'collect-logs' command in order to connect to Jethro and collect SQL statistics.${NORMAL}"
            noSqlStats_YesNo
        else
            echo -e "${YELLOW}SQL statistics collected successfully.${NORMAL}"
        fi
    else
        echo -e "Service for instance ${RED}<$INSTANCE_NAME>${NORMAL} has not been properly registered."
        echo 'Could not get SQL statistics.'
        noSqlStats_YesNo
    fi

    parse_local_config
    mkdir -p $SESSION_TMP/metaData
        mkdir -p $SESSION_TMP/loads

    case $STORAGE_TYPE in
      "POSIX" )
        cp $INSTANCE_STORAGE_PATH/metaData/schema*xml $SESSION_TMP/metaData/
                cp $INSTANCE_STORAGE_PATH/loads/*xml $SESSION_TMP/loads/
        ;;
      "HDFS" )
        hadoop fs -copyToLocal $INSTANCE_STORAGE_PATH/metaData/schema*xml $SESSION_TMP/metaData/
                hadoop fs -copyToLocal $INSTANCE_STORAGE_PATH/loads/*xml $SESSION_TMP/loads/
        ;;
    esac

    tar -cvhf $JETHRO_TAR -C /tmp jethro_diagnostics.out -C $SESSION_TMP dmesg.out crontab.out services.ini sqlstatistics.out metaData/ loads/ -C $LOGS monitor.log $INSTANCE_NAME -C $CONF $INSTANCE_NAME/local-conf.ini >/dev/null 2>&1
  fi

  mv $JETHRO_TAR ..
  cd ..
  #Free up space of our temp files -- everything is inside $JETHRO_TAR already
  rm -rf $SESSION_TMP/ /tmp/jethro_diagnostics.out >/dev/null 2>&1
  gzip -9f $JETHRO_TAR >/dev/null 2>&1

  # other temp files will be removed by on_exit handler
  echo -e "${YELLOW}DONE. The stats are located in /tmp/$JETHRO_TAR.gz${NORMAL}"

  exit 0
fi

check_jethro_home

if [[ $OPERATION == create-instance ]]; then
    #validating instance name length
    if (( ${#INSTANCE_NAME} > $MAX_NAME_LEN)); then
      echo -e "Error, invalid instance name ${RED}<$2>${NORMAL}, aborting operation."
      echo "Note: instance name length should not exceed $MAX_NAME_LEN chars"
      exit 13
    fi
    #Checking if the Instance-name is in the reserve words array {RESERVED_WORDS}
    reservedWordsArray=$(echo ${RESERVED_WORDS[@]} | grep -w "${INSTANCE_NAME,,}" |  wc -w)
    if [ $reservedWordsArray -ne 0 ] ; then
      echo
      echo -e "Error - Instance name $RED$INSTANCE_NAME$NORMAL is not allowed, please choose a different Instance name."
      exit 18
    fi

fi

if [[ $OPERATION == create-instance || $OPERATION == attach-instance ]]; then
    if [[ "$@" != *"-storage-path"* ]]; then
       echo "ERROR - missing storage root path parameter, aborting operation."
       exit 12
    fi
    #add validation for -cache and -size
    if [[ "$@" != *"-cache-path"* ]]; then
     echo "ERROR - missing cache root path parameter, aborting operation."
     exit 12
    fi
    if [[ "$@" != *"-cache-size"* ]]; then
     echo "ERROR - missing cache max size parameter, aborting operation."
     exit 12
    fi
fi

if [[ $OPERATION == list-storage-instances ]]; then
    if [[ "$@" != *"-storage-path"* ]]; then
       echo "ERROR - missing storage root path parameter, aborting operation."
       exit 99
    fi
fi
#Check if set-host-config came with a parameter
if [[ $OPERATION == set-host-config ]]; then
    if [[ "$#" -lt 2 ]]; then
       echo "ERROR - At least one parameter should be configured, aborting operation."
       exit 99
    fi
fi

case $OPERATION in

   set-host-config)
   #################
   # SET-HOST-CONFIG
   #################
   if [[ "$#" -gt 1 ]]; then
        EXTRA_PARAMS=${@:2} #copy parameters. Starting from the second parameters
        checkingCorrectHostConfParameters
        check_extra_options
   fi
;;

   create-instance)
   #################
   # CREATE INSTANCE
   #################

     STORAGE_TYPE=HDFS
     if [[ "$#" -gt 3 ]]; then
       EXTRA_PARAMS=${@:3}
       check_extra_options
     fi
     if [[ $DEBUG == TRUE ]]; then echo "  DEBUG: STORAGE_TYPE $STORAGE_TYPE" >> /tmp/jethro_diagnostics.out 2>&1 ; fi

     check_local_dir
     if [[ $LOCAL_DIR_EXISTS != "FALSE" ]]; then
       echo
       echo -e "Error - instance ${RED}<$INSTANCE_NAME>${NORMAL} already exists, aborting operation."
       echo "Reason - found local directory $JETHRO_HOME/instances/$INSTANCE_NAME"
       exit 12
     fi

     #STORAGE_ROOT_PATH=$3
     INSTANCE_STORAGE_PATH=$STORAGE_ROOT_PATH"/"$INSTANCE_NAME
     # check central storage (HDFS/POSIX)
     if [[ $STORAGE_TYPE == "HDFS" ]]; then
        check_hdfs_dir
        if [[ $HDFS_DIR_EXISTS != "FALSE" ]]; then
          echo
          echo -e "Error - instance ${RED}<$INSTANCE_NAME>${NORMAL} already exists on HDFS, aborting operation."
          echo "Reason - found storage directory $INSTANCE_STORAGE_PATH on HDFS"
          echo "Consider attaching to that instance instead by running:"
          generate_attach_statement
          exit 14
        else
          CREATE_STORAGE_FUNC=create_hdfs_instance
        fi
     else # POSIX
        check_posix_dir
        if [[ $POSIX_DIR_EXISTS != "FALSE" ]]; then
          echo
          echo -e "Error - instance ${RED}<$INSTANCE_NAME>${NORMAL} already exists locally, aborting operation."
          echo "Reason - found storage directory $INSTANCE_STORAGE_PATH locally"
          echo "Consider attaching to that instance instead by running:"
          generate_attach_statement
          exit 14
        else
          if [[ ! -d $STORAGE_ROOT_PATH ]]; then
            echo -e "Error - parameter storage root path directory ${RED}<$STORAGE_ROOT_PATH>${NORMAL} does not exist, aborting operation."
            exit 18
          fi
          if [[ ! -w $STORAGE_ROOT_PATH ]]; then
            echo -e "Error - parameter storage root path directory ${RED}<$STORAGE_ROOT_PATH>${NORMAL} is not writable, aborting operation."
            exit 18
          fi
        fi
        CREATE_STORAGE_FUNC=create_posix_instance
     fi
     AUTOSTART_MAINT_SCHEDULER="yes"
     create_local_instance
     $CREATE_STORAGE_FUNC
         echo
     echo "Instance $INSTANCE_NAME was created successfully"
     #print_autostart_message
     $SERVICE jethro start $INSTANCE_NAME
     $SERVICE jethro start $INSTANCE_NAME maint
     $SERVICE jethro start $INSTANCE_NAME loadscheduler
         create_schemas
     create_system_tables
     ;;

   delete-instance)
   #################
   # DELETE_INSTANCE
   #################
     if [[ $# -gt 2 ]]; then
       echo "Error - too many parameters for delete-instance, aborting operation."
       echo
       display_usage
     fi

     check_local_dir
     if [[ $LOCAL_DIR_EXISTS != "TRUE" ]]; then
       echo
       echo -e "Error - instance ${RED}<$INSTANCE_NAME>${NORMAL} does not exist, aborting operation."
       echo "Reason - could not find local directory $JETHRO_HOME/instances/$INSTANCE_NAME"
       exit 12
     fi

     parse_local_config

     # check central storage (HDFS/POSIX)
     if [[ $STORAGE_TYPE == "HDFS" ]]; then
        check_hdfs_dir
        if [[ $HDFS_DIR_EXISTS != "TRUE" ]]; then
          echo "Warning - could not find storage directory $INSTANCE_STORAGE_PATH on HDFS, but local directories exist. Continuing will delete them. "
          DELETE_STORAGE_FUNC=
        else
            DELETE_STORAGE_FUNC=delete_hdfs_instance
        fi
     else # POSIX
        check_posix_dir
        if [[ $POSIX_DIR_EXISTS != "TRUE" ]]; then
          echo "Warning - could not find storage directory $INSTANCE_STORAGE_PATH locally, but local directories exist. Continuing will delete them. "
          DELETE_STORAGE_FUNC=
        else
          DELETE_STORAGE_FUNC=delete_posix_instance
        fi
      fi

      while true; do
        read -p "About to delete instance <$INSTANCE_NAME> - ALL DATA WILL BE LOST. Are you sure you want to delete this instance? [no]?" ANSWER

        ANSWER=`echo ${ANSWER} | tr [:upper:] [:lower:]`
        ANSWER=${ANSWER:-no}

        case $ANSWER in
        yes )
          # delete from storage before local - so if something fails, can run again (local directory will still exist)
          echo
          $DELETE_STORAGE_FUNC
          delete_local_instance
          break;;
        no )
          echo "Exiting..."
          exit 0;;
         * )
          echo "Please answer yes or no.";;
        esac
      done

      echo
      echo "Instance $INSTANCE_NAME was deleted successfully"
    ;;
   set-host-config)
   #################
   # SET HOST CONFIG
   #################

     if [[ $# -lt 1 ]]; then
       echo "Error - At least one parameter should be set, aborting operation."
       echo
        display_usage
     fi
   ;;


   attach-instance)
   #################
   # ATTACH INSTANCE
   #################

     STORAGE_TYPE=HDFS
     if [[ "$#" -gt 3 ]]; then
       EXTRA_PARAMS=${@:3}
       check_extra_options
     fi
     if [[ $DEBUG == TRUE ]]; then echo "  DEBUG: STORAGE_TYPE $STORAGE_TYPE" >> /tmp/jethro_diagnostics.out 2>&1 ; fi

     check_local_dir
     if [[ $LOCAL_DIR_EXISTS != "FALSE" ]]; then
       echo
       echo -e "Error - instance ${RED}<$INSTANCE_NAME>${NORMAL} already attached, aborting operation."
       echo "Reason - found local directory $JETHRO_HOME/instances/$INSTANCE_NAME"
       if [[ $STORAGE_TYPE == "HDFS" ]]; then
           print_unattached_hdfs_instances
       else
           print_unattached_posix_instances
       fi
       exit 12
     fi

     INSTANCE_STORAGE_PATH=$STORAGE_ROOT_PATH"/"$INSTANCE_NAME
     # check central storage (HDFS/POSIX)
     if [[ $STORAGE_TYPE == "HDFS" ]]; then
        check_hdfs_dir
        if [[ $HDFS_DIR_EXISTS != "TRUE" ]]; then
          echo -e "Error - instance ${RED}<$INSTANCE_NAME>${NORMAL} does not exist, aborting operation."
          echo "Reason - could not find directory $INSTANCE_STORAGE_PATH on HDFS"
          print_unattached_hdfs_instances
          exit 99
        fi
     else # POSIX
        check_posix_dir
        if [[ $POSIX_DIR_EXISTS != "TRUE" ]]; then
          echo -e "Error - instance ${RED}<$INSTANCE_NAME>${NORMAL} does not exist, aborting operation."
          echo "Reason - could not find  directory $INSTANCE_STORAGE_PATH on local storage"
          print_unattached_posix_instances
          exit 99
        fi
      fi

     AUTOSTART_MAINT_SCHEDULER="no"
     create_local_instance

     echo
     echo "Instance $INSTANCE_NAME was attached successfully"
     #print_autostart_message
     $SERVICE jethro start $INSTANCE_NAME
     ;;

   detach-instance)
   #################
   # DETACH INSTANCE
   #################

     if [[ $# -gt 2 ]]; then
       echo "Error - too many parameters for detach-instance, aborting operation."
       echo
        display_usage
     fi

     check_local_dir
     if [[ $LOCAL_DIR_EXISTS != "TRUE" ]]; then
       echo
       echo -e "Error - instance ${RED}<$INSTANCE_NAME>${NORMAL} does not exist, aborting operation."
       echo -e "Reason - could not find local directory $JETHRO_HOME/instances/$INSTANCE_NAME"
           echo
       FOUND_INSTANCES=false
       for INSTANCES in `ls -d $JETHRO_HOME/instances/*/ 2> /dev/null` ; do
         if [[ $FOUND_INSTANCES != true ]]; then
           FOUND_INSTANCES=true
           echo -e "${YELLOW}There are other JethroData instances currently attached to this server:"${GREEN}
         fi
         echo -e "  " `basename $INSTANCES`
       done
       echo -en $NORMAL
       exit 12
     fi

     parse_local_config

     # check central storage (HDFS/POSIX)
     if [[ $STORAGE_TYPE == "HDFS" ]]; then
        check_hdfs_dir
        if [[ $HDFS_DIR_EXISTS != "TRUE" ]]; then
          echo "Warning - could not find storage directory $INSTANCE_STORAGE_PATH on HDFS, but local directories exist - they will be removed. "
        fi
     else # POSIX
        check_posix_dir
        if [[ $POSIX_DIR_EXISTS != "TRUE" ]]; then
          echo "Warning - could not find storage directory $INSTANCE_STORAGE_PATH locally, but local directories exist - they will be removed. "
        fi
     fi

     delete_local_instance
     echo
     echo "Instance $INSTANCE_NAME was detached successfully"
     echo
     echo "To attach again to this instance, run:"
     generate_attach_statement

   ;;

   reset-instance)
   #################
   # RESET INSTANCE
   #################
     if [[ $# -gt 2 ]]; then
       echo "Error - too many parameters for reset-instance, aborting operation."
       echo
       display_usage
     fi

     check_local_dir
     if [[ $LOCAL_DIR_EXISTS != "TRUE" ]]; then
       echo
       echo -e "Error - instance ${RED}<$INSTANCE_NAME>{$NORMAL} does not exist, aborting operation."
       echo "Reason - could not find local directory $JETHRO_HOME/instances/$INSTANCE_NAME"
       exit 12
     fi

     parse_local_config

     # check central storage (HDFS/POSIX)
     if [[ $STORAGE_TYPE == "HDFS" ]]; then
        check_hdfs_dir
        if [[ $HDFS_DIR_EXISTS != "TRUE" ]]; then
          echo "Warning - could not find storage directory $INSTANCE_STORAGE_PATH on HDFS, but local directories exist. Continuing will skip HDFS. "
          RESET_STORAGE_FUNC=
        else
          RESET_STORAGE_FUNC=reset_hdfs_instance
        fi
     else # POSIX
        check_posix_dir
        if [[ $POSIX_DIR_EXISTS != "TRUE" ]]; then
          echo "Warning - could not find storage directory $INSTANCE_STORAGE_PATH locally, but local directories exist. Continuing will skip local storage. "
          RESET_STORAGE_FUNC=
        else
          RESET_STORAGE_FUNC=reset_posix_instance
        fi
      fi

      while true; do
        read -p "About to reset instance ${RED}<$INSTANCE_NAME>${NORMAL} - ALL DATA WILL BE LOST. Are you sure you want to delete this instance? [no]?" ANSWER

        ANSWER=`echo ${ANSWER} | tr [:upper:] [:lower:]`
        ANSWER=${ANSWER:-no}

        case $ANSWER in
          yes )
            echo
            $RESET_STORAGE_FUNC
            delete_cache
            break;;
          no )
            echo "Exiting..."
            exit 0;;
          * )
            echo "Please answer yes or no.";;
        esac
      done

      echo
      echo "Instance $INSTANCE_NAME was reset successfully"

   ;;

   check-env)
   #################
   # Post install check
   #################
   check_local_dir

   if [[ $STORAGE_TYPE == 'POSIX' ]]; then
      check_posix_dir
   elif [[ $STORAGE_TYPE == 'HDFS' ]]; then
      check_hdfs_dir
   fi
   if [[ ! $CLASSPATH ]]; then
      echo
      echo "Error - \$CLASSPATH is not defined"
      exit 10
   fi

   echo
   echo 'Environment is Ok '
   ;;


   list-instances)
   #################
   # LIST INSTANCES
   #################
     if [[ $# -gt 1 ]]; then
       echo "Error - too many parameters for list-instances, aborting operation."
       echo
       display_usage
     fi

     check_local_dir
     get_local_instances

     if [[ "${#LOCAL_LIST[@]}" != 0 ]]; then
        echo -e ${YELLOW}"The following JethroData instances are currently attached to this server:"${GREEN}
        for inst in ${LOCAL_LIST[@]}; do
            echo '    ' $inst
        done
        echo -ne ${NORMAL}
     else
        echo -e ${YELLOW}"There are no JethroData instances currently attached to this server"${NORMAL}
     fi
   ;;


   list-storage-instances)
   ##########################
   # LIST STORAGE INSTANCES
   ##########################
     if [[ $# -gt 3 ]]; then
       echo "Error - too many parameters for list-storage-instances, aborting operation."
       echo
       display_usage
     fi

     EXTRA_PARAMS=${@:2}
     check_extra_options

     check_local_dir
     get_local_instances

     if [[ $STORAGE_TYPE == "HDFS" ]]; then
        check_hdfs_root
        get_hdfs_instances
        STORAGE_INSTANCES_LIST=("${HDFS_INSTANCES_LIST[@]}")
        echo
     else
        check_posix_root
        get_posix_instances
        STORAGE_INSTANCES_LIST=("${POSIX_INSTANCES_LIST[@]}")
        echo
     fi

     echo -en ${YELLOW}
     if [ ${#STORAGE_INSTANCES_LIST[@]} -eq 0 ]; then
         echo "There are no Jethro instances located under "${STORAGE_ROOT_PATH}
     else
         echo "Following Jethro instances are located under "${STORAGE_ROOT_PATH}
     fi
     echo -en ${NORMAL}

     #Find max instance name length
     ((max_output_len=0))
     for instance in "${STORAGE_INSTANCES_LIST[@]}"; do
         inst_len=${#instance}
         max_output_len=$((inst_len>max_output_len ? inst_len : max_output_len ))  #max
     done

     # output no more than 32 chars of instance name
     max_output_len=$((MAX_NAME_LEN>max_output_len ? max_output_len : MAX_NAME_LEN))
     for instance in "${STORAGE_INSTANCES_LIST[@]}"; do
         test -z "${LOCAL_LIST[$instance]}" && msg='Not attached' || msg='Attached'
         inst_len=${#instance}
         output_len=$((inst_len>max_output_len ? max_output_len : inst_len))        #truncate instance name to max_output_len chars
         output_name=${instance:0:output_len}
         printf " %-"$max_output_len"s | %s" "$output_name" "$msg"
         echo
     done
   ;;


   change-pass)
   #################
   # CHANGE PASSWORD
   #################
     if [[ $# -gt 2 ]]; then
       echo "Error - too many parameters for change-pass, aborting operation."
       echo
       display_usage
     fi

     check_local_dir
     if [[ $LOCAL_DIR_EXISTS != "TRUE" ]]; then
       echo
       echo -e "Error - instance ${RED}<$INSTANCE_NAME>${NORMAL} does not exist, aborting operation."
       echo "Reason - could not find local directory $JETHRO_HOME/instances/$INSTANCE_NAME"
       exit 12
     fi

     Jethro P $INSTANCE_NAME

     #add vadidation for change pass!
   ;;

   # else - unkown operation
   *)
     display_usage
;;
esac

exit 0
